#include "../StdAfx.h"
#include "TrawMesh.h"
#include "../GraphicsAPI/Tmaterial.h"

static Tmaterial sfToShade;

TrawMesh::TrawMesh() : fObjectCenter(0, 0, 0)
{
	fMesh = NULL;
	fMeshMaterials = NULL;
	fMeshTextures  = NULL;
	fNumMaterials = 0;
	fDevice = NULL;
	fStack = NULL;
	fObjectRadius = 1.0;


	sfToShade.SetAmbient(0, 0, 0, 0);
	sfToShade.SetDiffuse(0, 0, 0, 0);
	sfToShade.SetSpecular(0, 0, 0, 0);
	sfToShade.SetEmissive(0, 0, 0, 0);
	sfToShade.SetPower(0);
}

TrawMesh::~TrawMesh() 
{
	if (fDevice != NULL) {

		if( fMeshMaterials != NULL ) 
			delete[] fMeshMaterials;

		if( fMeshTextures != NULL) {

			for( unsigned int i = 0; i < fNumMaterials; i++ ) {
				if( fMeshTextures[i] != NULL )
					fMeshTextures[i]->Release();
			}
			delete[] fMeshTextures;
		}

		if( fMesh != NULL )
			fMesh->Release();
	}
}


//-----------------------------------------------------------------------------
// Load the mesh (in D3D .x format) and store into the buffer of the device
//
BOOL TrawMesh::LoadMesh(
	LPDIRECT3DDEVICE9 pDevice,
	ID3DXMatrixStack* stack, 
	CString			  path,
	CString			  fileName
)
{
    LPD3DXBUFFER pD3DXMtrlBuffer;
	CString		 fullPathName;

	fDevice = pDevice;
	fStack = stack;

    // Load the mesh from the specified file
	fullPathName.Format("%s/%s", path, fileName);
    if( FAILED( D3DXLoadMeshFromX( fullPathName.GetBuffer(0), D3DXMESH_SYSTEMMEM, 
                                   fDevice, NULL, 
                                   &pD3DXMtrlBuffer, NULL, &fNumMaterials, 
                                   &fMesh ) ) )
    {
        return FALSE;
    }

    // We need to extract the material properties and texture names from the 
    // pD3DXMtrlBuffer
    D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)pD3DXMtrlBuffer->GetBufferPointer();
    fMeshMaterials = new D3DMATERIAL9[fNumMaterials];
    fMeshTextures  = new LPDIRECT3DTEXTURE9[fNumMaterials];

    for( unsigned int i=0; i<fNumMaterials; i++ )
    {
        // Copy the material
        fMeshMaterials[i] = d3dxMaterials[i].MatD3D;

        // Set the ambient color for the material (D3DX does not do this)
        // fMeshMaterials[i].Ambient = fMeshMaterials[i].Diffuse;
		//
		// For now, since we do not have light source, let the object emit energy
		// fMeshMaterials[i].Emissive = fMeshMaterials[i].Diffuse;
     

        // Create the texture
		fullPathName.Format("%s/%s", path, d3dxMaterials[i].pTextureFilename);
        if( FAILED( D3DXCreateTextureFromFile( fDevice, 
                                               fullPathName,
                                               &fMeshTextures[i] ) ) )
        {
            fMeshTextures[i] = NULL;
        }
    }

    // Done with the material buffer
    pD3DXMtrlBuffer->Release();

	//
	// now compute the bounding sphere of this mesh 

	LPDIRECT3DVERTEXBUFFER9 pVertexBuffer = NULL;
	void*        pVertices;
    // Lock the vertex buffer, to generate a simple bounding sphere
    fMesh->GetVertexBuffer( &pVertexBuffer );
    pVertexBuffer->Lock( 0, 0, &pVertices, D3DLOCK_NOSYSLOCK );
   
    D3DXComputeBoundingSphere( (D3DXVECTOR3*)pVertices, fMesh->GetNumVertices(),
                                    D3DXGetFVFVertexSize(fMesh->GetFVF()),
                                    &fObjectCenter, &fObjectRadius );
    pVertexBuffer->Unlock();
	pVertexBuffer->Release();

	fObjectCenter = -fObjectCenter;
	fObjectRadius = 1.0/fObjectRadius;

	// #ifdef _IF_TEST_INDEXBUFFER_
	{
		//
		// Try getting at the IndexBuffer ...
		//
		DWORD f = fMesh->GetFVF();
		int size = D3DXGetFVFVertexSize(f);

		LPDIRECT3DINDEXBUFFER9 pIB;
		fMesh->GetIndexBuffer( &pIB );
		int numFaces = fMesh->GetNumFaces();
		void *pBuffer;
		pIB->Lock(0, 0, &pBuffer, D3DLOCK_NOSYSLOCK);
			unsigned short *IndexArray = (unsigned short *)pBuffer;
		pIB->Unlock();
	}
	// #endif


    return TRUE;
}

VOID TrawMesh::SetUpD3DStack(ID3DXMatrixStack *stack) 
{
	stack->ScaleLocal(fObjectRadius, fObjectRadius, fObjectRadius);
	stack->TranslateLocal(fObjectCenter.x, fObjectCenter.y, fObjectCenter.z);
}
VOID TrawMesh::DrawRawMesh(bool shouldLight) 
{
	// Meshes are divided into subsets, one for each material. Render them in
	// a loop
	for( unsigned int i=0; i<fNumMaterials; i++ )	{

		// Set the material and texture for this subset
		if (shouldLight)
			fDevice->SetMaterial( &fMeshMaterials[i] );
	    

		fDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);

		if (fMeshTextures[i] != NULL) {
				//
				// set up texture to be the color of the mesh
				//
				fDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
  				fDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);

				fDevice->SetTexture( 0, fMeshTextures[i] );
		} 

	    
		// Draw the mesh subset
		fMesh->DrawSubset( i );
	}
	//
	// Switch off texture
	//
	fDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_DISABLE);
}

//-----------------------------------------------------------------------------
VOID TrawMesh::DrawMesh(bool shouldLight)
{
    fStack->Push();
		SetUpD3DStack(fStack);

		fDevice->SetTransform(D3DTS_WORLD, fStack->GetTop());
		DrawRawMesh(shouldLight);
		
	fStack->Pop();
	fDevice->SetTransform(D3DTS_WORLD, fStack->GetTop());
}
