#ifndef TLIGHT_H
#define TLIGHT_H

#include "GraphicsAPI.h"
#include "TdrawableObject.h"

#define SetLightColor(materialType) 									\
	void Set##materialType##(float v)	{								\
		fLt.##materialType##.r = v;										\
		fLt.##materialType##.g = v;										\
		fLt.##materialType##.b = v;										\
		fLt.##materialType##.a = 1.0;									\
	}																	\
	void Set##materialType##(float r, float g, float b, float a=1.0) {	\
		fLt.##materialType##.r = r;										\
		fLt.##materialType##.g = g;										\
		fLt.##materialType##.b = b;										\
		fLt.##materialType##.a = a;										\
	}																	\
	void Set##materialType##(const D3DXCOLOR &v) {						\
		fLt.##materialType##.r = v.r;									\
		fLt.##materialType##.g = v.g;									\
		fLt.##materialType##.b = v.b;									\
		fLt.##materialType##.a = v.a;									\
	}																	\
	void Get##materialType##(D3DXCOLOR &v) const {						\
		v.r = fLt.##materialType##.r;									\
		v.g = fLt.##materialType##.g;									\
		v.b = fLt.##materialType##.b;									\
		v.a = fLt.##materialType##.a;									\
	}


#define SetLightPosOrDir(_posOrDir_) 									\
	void Set##_posOrDir_##(float v)	{									\
		fLt.##_posOrDir_##.x = v;										\
		fLt.##_posOrDir_##.y = v;										\
		fLt.##_posOrDir_##.z = v;										\
	}																	\
	void Set##_posOrDir_##(float x, float y, float z) {					\
		fLt.##_posOrDir_##.x = x;										\
		fLt.##_posOrDir_##.y = y;										\
		fLt.##_posOrDir_##.z = z;										\
	}																	\
	void Set##_posOrDir_##(const D3DXVECTOR3 &v) {						\
		fLt.##_posOrDir_##.x = v.x;										\
		fLt.##_posOrDir_##.y = v.y;										\
		fLt.##_posOrDir_##.z = v.z;										\
	}																	\
	void Get##_posOrDir_##(D3DXVECTOR3 &v) const {						\
		v.x = fLt.##_posOrDir_##.x;										\
		v.y = fLt.##_posOrDir_##.y;										\
		v.z = fLt.##_posOrDir_##.z;										\
	}


#define SetLightAttr(_attr_) 											\
	void Set##_attr_##(float v) {										\
		fLt.##_attr_## = v;												\
	}																	\
	float Get##_attr_##() {												\
		return fLt.##_attr_##;											\
	}

//
// light information for D3DX: We wrap around D3DLIGHT9 structure providing a collection
// of accessing methods.
//
class Tlight : public TdrawableObject {
	//
	// Notice: we are again sub-classing from TdrawableObject
	//         because we want to draw the lights
	//
	//         We have to watch out that the position of the light is stored both in 
	//         TgraphicsObject::XformInfo and in fLt.Position
	//
	//         We will use -y-axis in the TgraphicsObject::fXformInfo::fRotationMatrix
	//         to represent our fLt.Direction
	//         
	//         The light color (Ambient/Diffuse/Specular) are stored in Tmaterial also
	//         ui will set the Tmaterial, and before we set D3DX light we must get
	//         our light color from Tmaterial
	//
	private:
		int			fLightID;	//
								// There are about 8-16 hardware light sources, use with care!!
								//

		D3DLIGHT9	fLt;		// this is the light
								// fLt has: 	Type		(D3DLIGHT_POINT, DIRECTIONAL, or SPOT)
								// 				Ambient		(D3DCOLORVALUE)
								// 				Diffuse		(D3DCOLORVALUE)
								// 				Specular	(D3DCOLORVALUE)
								// 				Position	(D3DVECTOR)
								// 				Direction	(D3DVECTOR)
								// 				Range		float
								// 				Falloff		float
								// 				Attenuation0, 1, 2
								// 				Theta		(for spot light)
								// 				Phi


		//
		// Outside world should change the light position/direction 
		// from TgraphicsObject::fXformInfo
		SetLightPosOrDir(Position)
		SetLightPosOrDir(Direction)

	protected:
	
	public:

		Tlight(int id) {
			fLightID = id;

			SetToPointLight();

			// Nice set of default values ...
			// 
			SetDiffuse(0.8f);
			SetSpecular(0.2f);
			SetAmbient(0.1f);

			//
			// Set to the same as the above
			fMaterial.SetDiffuse(0.8f);
			fMaterial.SetSpecular(0.2f);
			fMaterial.SetAmbient(0.1f);
			
			SetPosition(0, 0, 0);
			SetDirection(0, -1, 0);
		
			SetRange(2000.0f);
			SetFalloff(1.0f);
			SetAttenuation0(1.0f);
			SetAttenuation1(0.0f);
			SetAttenuation2(0.0f);
			SetTheta(M_PI/4.0f);		// Spot light inner cone radius
			SetPhi(M_PI/3.0f);			// Spot light outter cone radius
		}

		D3DLIGHTTYPE GetLightType() {
			return fLt.Type;
		}
		void SetToSpotLight() {
			fLt.Type = D3DLIGHT_SPOT;
			SetPhi(M_PI/3.0);
			SetTheta(M_PI/4.0);
			SetFalloff(1.0);
		}
		void SetToPointLight() {
			fLt.Type = D3DLIGHT_POINT;
			SetPhi(0.0);
			SetTheta(0.0);
			SetFalloff(0.0);
			SetDirection(0, 0, 0);
		}
		void SetToDirectionalLight() {
			fLt.Type = D3DLIGHT_DIRECTIONAL;
			SetPhi(0.0);
			SetTheta(0.0);
			SetFalloff(0.0);
		}

		//
		// define the setting and accessing functions
		//
		SetLightColor(Ambient)
		SetLightColor(Diffuse)
		SetLightColor(Specular)
		
		SetLightAttr(Range)
		SetLightAttr(Falloff)
		SetLightAttr(Attenuation0)
		SetLightAttr(Attenuation1)
		SetLightAttr(Attenuation2)
		SetLightAttr(Theta)
		SetLightAttr(Phi)

		//
		// D3D Calls to switch on/off load light information
		//
		void SwitchOnLight(LPDIRECT3DDEVICE9 device) {
			TxformInfo &xt = GetXformInfo();
			//
			// 1. Update the fLt.Position/Direction from the TgraphicsObject::fXformInfo
			const D3DXMATRIX &m = xt.GetRotationMatrix();
			SetPosition(xt.Movement());
			SetDirection(-m._21, -m._22, -m._23);

			// 2. Update the fLt.Ambient/Diffuse/Specular from TgraphicsObject::fMaterial
			D3DXCOLOR ca, cd, cs;
			fMaterial.GetAmbient(ca);
			fMaterial.GetDiffuse(cd);
			fMaterial.GetSpecular(cs);
			SetAmbient(ca);
			SetDiffuse(cd);
			SetSpecular(cs);
			
			//
			// Now load the new fLt infomation into D3D and
			// switch on the light
			//
			device->SetLight(fLightID, &fLt);
			device->LightEnable(fLightID, TRUE);
		}

		void SwitchOffLight(LPDIRECT3DDEVICE9 device) {
			device->LightEnable(fLightID, FALSE);
		}

		virtual void GetLightDirection(D3DXVECTOR3 &dir) const {
			// direction light has dir.w = 0
			const D3DXMATRIX &m = fXformInfo.GetRotationMatrix();
			dir.x = -m._21;
			dir.y = -m._22;
			dir.z = -m._23;
		}

		virtual void GetLightPosition(D3DXVECTOR4 &pos) const {
			const D3DXVECTOR3 &p = this->GetLocation();
			pos.x = p.x;
			pos.y = p.y;
			pos.z = p.z;
			pos.w = 1.0;
		}
};
#endif

