
#include "TmyWorld.h"
#include "../Td3dDrawable.h"
#include "../Resource.h"
#include "../TmodelControl.h"

void TmyWorld::DrawMirror(Td3dDrawable &d)
{
	LPDIRECT3DDEVICE9 device = d.D3DDevice();
	ID3DXMatrixStack *stack = d.Obj2WldStack();

	// 1. Check if reflection is switched on
	if (fControl->m_DoReflection) {
	
		stack->Push();
		// save a copy of the top of stack ..

			// Get Mirror geometry ...
			//
			D3DXVECTOR3 n, p0, p1, p2, p3;	// normal vector and three vertices ...
			float		pd;					// d for the plane equation 
			fMirror.GetPlaneEquationAndVertices(n, pd, p0, p1, p2, p3);
			D3DXPLANE plane(n.x, n.y, n.z, pd);
			D3DXMATRIX mirrorXform;
			D3DXMatrixReflect( &mirrorXform, &plane);
			stack->MultMatrix(&mirrorXform );
	
			// 2. Check if stencil is switched on ...
			//
			if (fControl->m_UseStencil) {
				//
				// This is what we are going to do:
				//
				//
				// 1. We are going to "Paint" the "fMirror" position
				//    onto the stencil buffer. We don't want to see anything
				//    all we want here is to know "where is the mirror".
				// 2. After 1, we have a stencil buffer telling us where is
				//    the mirror, now when we render the reflected objects
				//    only at places where the mirror is present would we draw
				//
				//
				// To achieve 1: 
				// 		-> a. Switch on Stencil buffer (always write into it)
				// 		-> b. Switch off both Z-Buffer, and Frame Buffer
				//                -> no depth testing, and we DO NOT want to
				//				     see color appearing
				//      -> c. Now render the fMirror, and write "1" into all 
				//				stencil buffer positions touched.
				//
				// To achieve 2:
				//      -> Switch on z-buffer, and frame buffer
				//      -> render the reflected object, and ONLY write into frame buffer
				//             where stencil buffer has 1 in there.
				//

				// So let's go
				//
				// To do 1: compute the stencil of the Mirror ...
				//
					device->SetRenderState(D3DRS_STENCILENABLE, TRUE);			
								// enable stencil buffer
					
					device->SetRenderState(D3DRS_STENCILFUNC,   D3DCMP_ALWAYS);	
								// StencilFun is always PASS
					
					device->SetRenderState(D3DRS_STENCILPASS,   D3DSTENCILCAPS_REPLACE);	
								// When PASS, replace Stencil Buffer content with StencilRef 
								//    the next line says StencilRef is 0x01

#define kReflectValue     0x1
#define kReflectWriteMask 0x1

					device->SetRenderState(D3DRS_STENCILWRITEMASK,    kReflectWriteMask);			
								// Make sure we only touch the reflection bit
					
					device->SetRenderState(D3DRS_STENCILREF,    kReflectValue);			
								// value to write into the stencil is: what we define above

					//     . now Disable Z-buffer checking and frame buffer writting
					device->SetRenderState(D3DRS_ZENABLE,   	D3DZB_FALSE);	// disable z-buffer comparison
					device->SetRenderState(D3DRS_ZWRITEENABLE,   FALSE);		// disable z-buffer writing

					//     . now disable frame buffer writing ...
					device->SetRenderState(D3DRS_COLORWRITEENABLE, 0);			// do not write into frame buffer
					
					//
					//
					device->SetRenderState( D3DRS_CULLMODE,	D3DCULL_CCW );	
								// this is to make sure we only see one side of the mirror
								// when the camera is moved to the "back" of the mirror, we 
								// won't draw the mirror into the stencil buffer and so
								// no reflected object will show up
								//
								//
								//  *** TRY *** - comment out the above line and move your
								//			camera to behind the mirror with 
								//			do reflect and stencil on.
								//
					//
					// Now Draw the mirror geometry ...
					//   notice, this draw will only fill the stencil buffer
					//   all stencil buffer pixel positions where fMirror touched
					//   will be painted with "1"
					//
					// Since both Z and framebuffer are off,  NOTHING ELSE will be affected
					//
    				fMirror.DrawObject(d);
	
					//
					// Restore render state ...
					//
					device->SetRenderState( D3DRS_CULLMODE,	D3DCULL_NONE);		
					device->SetRenderState(D3DRS_ZENABLE,  D3DZB_TRUE);			// enable z-buffer comparison
					device->SetRenderState(D3DRS_ZWRITEENABLE,   D3DZB_TRUE);	// enable z-buffer writing
					device->SetRenderState(D3DRS_COLORWRITEENABLE, 				// enable all RGBA frame buffer write
												D3DCOLORWRITEENABLE_RED|
												D3DCOLORWRITEENABLE_GREEN|
												D3DCOLORWRITEENABLE_BLUE|
												D3DCOLORWRITEENABLE_ALPHA);			

				//
				// To do 2. 
				// 	Set up stencil for drawing of reflected objects ...
				//
					device->SetRenderState(D3DRS_STENCILFUNC,   D3DCMP_EQUAL);	
									// Allows writing into color buffer only when stencil buffer location is STENCILREF
									// the next line of code says our stencilRef is 1
									//
					device->SetRenderState(D3DRS_STENCILREF,    kReflectValue);		
									// stencil ref is ...

					device->SetRenderState(D3DRS_STENCILMASK,    kReflectWriteMask);			
									// only look at the relevant bit

					device->SetRenderState(D3DRS_STENCILPASS,   D3DSTENCILCAPS_KEEP);	
								// When PASS, do not change Stencil Buffer content 
			}
	
	    	// Render the scene (reflected by the mirror)
	    	RedrawAllObjects(d);

					
			//
			// Clean-up/restore-state if necessary ...
			//
			if (fControl->m_UseStencil) {
				//
				// disable stencil buffer
				//
				device->SetRenderState(D3DRS_STENCILENABLE, FALSE);		
				device->SetRenderState(D3DRS_STENCILMASK, 0XFFFFFFFF);
				device->SetRenderState(D3DRS_STENCILWRITEMASK, 0XFFFFFFFF);
			}

		stack->Pop();
		device->SetTransform( D3DTS_WORLD, stack->GetTop() );
		// restore the stack before continue ...
	}

	
    // Finally, render the mirror itself
    fMirror.DrawObject(d);
}
