
#include "TmyWorld.h"
#include "../Td3dDrawable.h"
#include "../Resource.h"
#include "../TmodelControl.h"

void TmyWorld::DrawGroundAndShadow(Td3dDrawable &d)
{
	LPDIRECT3DDEVICE9 device = d.D3DDevice();
	// For Stencil buffering, we want to:
	//
	// 1. Identify where the ground plane is (so that shadow drawing will
	//    only be done within the ground plane). This is just like reflection
	//    where all reflectec objects are drawn only where the mirror plane is 
	//    located.
	//
	// 2. To avoid multiple shadow polygon overlapping creating noise/errornous
	//    z-depth, we for each ground location, we only want to draw shadow
	//    objects once. We will use stencil buffer to help us achieve this
	//    by replacing the stencil buffer content each time we draw
	//    shadow object into the buffer, and we will test to make sure
	//    there are no shadow object in the buffer before we draw ...

	if (fControl->m_DoShadow && fControl->m_UseStencil) {

		// now we want to achieve 1: swich on stencil buffer before
		// we draw the ground, and the ground will always put a 1 into the
		// stencil buffer ...
		//
		device->SetRenderState(D3DRS_STENCILENABLE, TRUE);			
					// enable stencil buffer
					
		device->SetRenderState(D3DRS_STENCILFUNC,   D3DCMP_ALWAYS);	
					// StencilFunc is always PASS
					
		device->SetRenderState(D3DRS_STENCILPASS,   D3DSTENCILCAPS_REPLACE);	
					// When PASS, replace Stencil Buffer content with StencilRef 
					//    the next line says StencilRef is 0x01

#define kShadowValue			0x01
			// use the first bit ...

#define kShadowWriteMask		0x03
			// we reserve 2 bits for shaodw, ... we are going to increase the stencil values ...


		device->SetRenderState(D3DRS_STENCILWRITEMASK,    kShadowWriteMask);			
					// Make sure we only touch our bits 

		device->SetRenderState(D3DRS_STENCILREF, kShadowValue);			
					// value to write into the stencil is as defined above

		// now when we draw the ground plane, it will put "1" into corresponding
		// locations in the stencil buffer ...
	}

	//
	// We draw the ground first, to let shadow "blend" with the ground ...
	//
	fGround.DrawObject(d);
		

	if (fControl->m_DoShadow) {

		if (fControl->m_UseStencil) {
					// now to achieve 1 & 2:
					//
					// -> We only write into color buffer when stencil buffer is already 1
					// -> When we write into the color buffer, we will increase stencil buffer
					//      content by one, this way no other shadow drawing will draw into
					//      the frame buffer again ...
			device->SetRenderState(D3DRS_STENCILFUNC,   D3DCMP_EQUAL);	
						// Allows writing into color buffer only when stencil buffer location is STENCILREF
						// the next line of code says our stencilRef is 01

			device->SetRenderState(D3DRS_STENCILMASK,    kShadowWriteMask);			
						// we have to bit positions ... 
						//    

			device->SetRenderState(D3DRS_STENCILREF,    kShadowValue);		
						// condition says ... if equal to this then ...

			device->SetRenderState(D3DRS_STENCILPASS,   D3DSTENCILCAPS_INCRSAT);	
						// When PASS, increase stencil buffer content by 1
						// This makes sure no other shaodw objects will draw
						// into this frame buffer location

		}
		D3DXVECTOR3 n, p0, p1, p2, p3;
		float pd;
		fGround.GetPlaneEquationAndVertices(n, pd, p0, p1, p2, p3);
		pd += -0.01f;	// slightly above the plane
		D3DXPLANE shadowReceiver(n.x, n.y, n.z, pd);
		
		LPDIRECT3DDEVICE9 device = d.D3DDevice();	
	
			DrawShadowFromLight(d, shadowReceiver, fPointLgt);
			DrawShadowFromLight(d, shadowReceiver, fDirLgt);
			DrawShadowFromLight(d, shadowReceiver, fSpotLgt);
			
			//
			// Clean-up/restore-state if necessary ...
			//
		if (fControl->m_UseStencil) {
			device->SetRenderState(D3DRS_STENCILMASK, 0XFFFFFFFF);
			device->SetRenderState(D3DRS_STENCILWRITEMASK, 0XFFFFFFFF);
			device->SetRenderState(D3DRS_STENCILENABLE, FALSE);		
					// disable stencil buffer
		}
	}

		
}

void TmyWorld::DrawShadowFromLight(
	Td3dDrawable &d,
	const D3DXPLANE &receiver,
	const Tlight &lgt
)
{
	if (lgt.VisibilityIsEnabled() ) {
		LPDIRECT3DDEVICE9 device = d.D3DDevice();
		ID3DXMatrixStack *stack = d.Obj2WldStack();

		stack->Push();

			D3DXVECTOR4 pos;
			lgt.GetLightPosition(pos);
			D3DXMATRIX shadowMatrix;

			D3DXMatrixShadow(&shadowMatrix, &pos, &receiver);
			stack->MultMatrix(&shadowMatrix );
		
				#define DrawForShadow(obj)	{				\
					bool l = obj.LightingIsEnabled();		\
					bool b = obj.AlphaBlendingIsEnabled();	\
					obj.DisableLighting();					\
					obj.EnableAlphaBlending();				\
					obj.DrawObject(d);						\
					if (l) obj.EnableLighting();			\
					if (!b) obj.DisableAlphaBlending(); 	\
				}
					DrawForShadow(fToyCraft);
					DrawForShadow(fTeapot);

			stack->Pop();
	}
}
